@@ -73,8 +73,8 @@
     (if (and olayout
              (not (mismatch (layout-inherits olayout) new-inherits)))
         olayout
-        (make-layout :classoid (make-undefined-classoid name)
-                     :%flags +condition-layout-flag+
+        (make-layout (make-undefined-classoid name)
+                     :flags +condition-layout-flag+
                      :inherits new-inherits
                      :depthoid -1
                      :length (layout-length cond-layout)))))
@@ -389,6 +389,7 @@
    Condition types are classes, but (as allowed by ANSI and not as described in
    CLtL2) are neither STANDARD-OBJECTs nor STRUCTURE-OBJECTs. WITH-SLOTS and
    SLOT-VALUE may not be used on condition objects."
+  (check-designator name define-condition)
   (let* ((parent-types (or parent-types '(condition)))
          (layout (find-condition-layout name parent-types))
          (documentation nil)
@@ -477,7 +478,7 @@
            (error "unknown option: ~S" (first option)))))
 
       ;; Maybe kill docstring, but only under the cross-compiler.
-      #!+(and (not sb-doc) (host-feature sb-xc-host)) (setq documentation nil)
+      #+(and (not sb-doc) sb-xc-host) (setq documentation nil)
       `(progn
          ,@(when *top-level-form-p*
              ;; Avoid dumping uninitialized layouts, for sb-fasl::dump-layout
@@ -637,16 +638,20 @@
   ((not-yet-loaded :initform nil :reader not-yet-loaded :initarg :not-yet-loaded))
   (:report
    (lambda (condition stream)
-     (let ((*package* (find-package :keyword)))
+     (let ((name (cell-error-name condition)))
        (format stream
-               "~@<The function ~S is undefined.~@?~@:>"
-               (cell-error-name condition)
+               (if (and (symbolp name) (macro-function name))
+                   (sb-format:tokens "~@<~/sb-ext:print-symbol-with-prefix/ is a macro, ~
+                                      not a function.~@:>")
+                   (sb-format:tokens "~@<The function ~/sb-ext:print-symbol-with-prefix/ ~
+                                      is undefined.~@?~@:>"))
+               name
                (case (not-yet-loaded condition)
                  (:local
-                  "~:@_It is a local function ~
-                       not available at compile-time.")
-                 ((t) "~:@_It is defined earlier in the ~
-                           file but is not available at compile-time.")
+                  (sb-format:tokens "~:@_It is a local function ~
+                                     not available at compile-time."))
+                 ((t) (sb-format:tokens "~:@_It is defined earlier in the ~
+                                         file but is not available at compile-time."))
                  (t
                   "")))))))
 
@@ -1084,12 +1089,26 @@ SB-EXT:PACKAGE-LOCKED-ERROR-SYMBOL."))
       (list '(:ansi-cl :function adjust-array))))
 
 (define-condition index-too-large-error (type-error)
-  ()
+  ((sequence :initarg :sequence))
   (:report
    (lambda (condition stream)
-     (format stream
-             "The index ~S is too large."
-             (type-error-datum condition)))))
+     (let ((sequence (slot-value condition 'sequence))
+           (index (type-error-datum condition)))
+       (if (vectorp sequence)
+           (format stream "Invalid index ~W for ~S ~@[with fill-pointer ~a~], ~
+                           should be a non-negative integer below ~W."
+                   index
+                   (type-of sequence)
+                   (and (array-has-fill-pointer-p sequence)
+                        (fill-pointer sequence))
+                   (length sequence))
+           (format stream
+                   "The index ~S is too large for a ~a of length ~s."
+                   index
+                   (if (listp sequence)
+                       "list"
+                       "sequence")
+                   (length sequence)))))))
 
 (define-condition bounding-indices-bad-error (reference-condition type-error)
   ((object :reader bounding-indices-bad-object :initarg :object))
@@ -1178,7 +1197,10 @@ SB-EXT:PACKAGE-LOCKED-ERROR-SYMBOL."))
 
 (define-condition simple-package-error (simple-condition package-error) ())
 
+(define-condition package-does-not-exist (simple-package-error) ())
+
 (define-condition simple-reader-package-error (simple-reader-error package-error) ())
+(define-condition reader-package-does-not-exist (simple-reader-package-error package-does-not-exist) ())
 
 (define-condition reader-eof-error (end-of-file)
   ((context :reader reader-eof-error-context :initarg :context))
@@ -1376,14 +1398,12 @@ handled by any other handler, it will be muffled.")
 ;;;; Deciding which redefinitions are "interesting".
 
 (defun function-file-namestring (function)
-  #!+sb-eval
-  (when (typep function 'sb-eval:interpreted-function)
+  (when (typep function 'interpreted-function)
     (return-from function-file-namestring
+      #+sb-eval
       (sb-c:definition-source-location-namestring
-          (sb-eval:interpreted-function-source-location function))))
-  #!+sb-fasteval
-  (when (typep function 'sb-interpreter:interpreted-function)
-    (return-from function-file-namestring
+          (sb-eval:interpreted-function-source-location function))
+      #+sb-fasteval
       (awhen (sb-interpreter:fun-source-location function)
         (sb-c:definition-source-location-namestring it))))
   (let* ((fun (%fun-fun function))
@@ -1403,9 +1423,7 @@ handled by any other handler, it will be muffled.")
           (typep new '(not compiled-function)))
      ;; fin->regular is interesting except for interpreted->compiled.
      (and (typep new '(not funcallable-instance))
-          (typep old '(and funcallable-instance
-                       #!+sb-fasteval (not sb-interpreter:interpreted-function)
-                       #!+sb-eval (not sb-eval:interpreted-function))))
+          (typep old '(and funcallable-instance (not interpreted-function))))
      ;; different file or unknown location is interesting.
      (let* ((old-namestring (function-file-namestring old))
             (new-namestring (function-file-namestring new)))
@@ -1502,6 +1520,10 @@ the usual naming convention (names like *FOO*) for special variables"
     (dubious-asterisks-around-variable-name)
   ())
 
+(define-condition &optional-and-&key-in-lambda-list
+    (style-warning simple-condition)
+  ())
+
 ;; We call this UNDEFINED-ALIEN-STYLE-WARNING because there are some
 ;; subclasses of ERROR above having to do with undefined aliens.
 (define-condition undefined-alien-style-warning (style-warning)
@@ -1510,7 +1532,7 @@ the usual naming convention (names like *FOO*) for special variables"
              (format stream "Undefined alien: ~S"
                      (undefined-alien-symbol warning)))))
 
-#!+(or sb-eval sb-fasteval)
+#+(or sb-eval sb-fasteval)
 (define-condition lexical-environment-too-complex (style-warning)
   ((form :initarg :form :reader lexical-environment-too-complex-form)
    (lexenv :initarg :lexenv :reader lexical-environment-too-complex-lexenv))
@@ -1678,9 +1700,7 @@ conditions."))
                             (deprecation-condition-software condition)
                             (deprecation-condition-name condition)))))))
 
-  ;; PRINT-SYMBOL-WITH-PREFIX is spelled using its target package name,
-  ;; not its cold package name, because these methods aren't unsable until
-  ;; warm load. (!CALL-A-METHOD does not understand method qualifiers)
+  ;; These conditions must not occur in self-build!
   (define-deprecation-warning early-deprecation-warning style-warning nil
      "~%~@<~:@_In future~@[ ~A~] versions ~
       ~/sb-ext:print-symbol-with-prefix/ will signal a full warning ~
@@ -1825,12 +1845,21 @@ the restart does not exist."))
 (define-condition simple-control-error (simple-condition control-error) ())
 (define-condition simple-file-error    (simple-condition file-error)    ())
 
+(defun %file-error (pathname &optional datum &rest arguments)
+  (typecase datum
+    (format-control (error 'simple-file-error :pathname pathname
+                                              :format-control datum
+                                              :format-arguments arguments))
+    (t (apply #'error datum :pathname pathname arguments))))
+
 (define-condition file-exists (simple-file-error) ())
 (define-condition file-does-not-exist (simple-file-error) ())
 (define-condition delete-file-error (simple-file-error) ())
 
-(define-condition simple-stream-error  (simple-condition stream-error)  ())
-(define-condition simple-parse-error   (simple-condition parse-error)   ())
+(define-condition simple-stream-error (simple-condition stream-error) ())
+(define-condition simple-parse-error  (simple-condition parse-error)  ())
+
+(define-condition broken-pipe (simple-stream-error) ())
 
 (define-condition character-coding-error (error)
   ((external-format :initarg :external-format :reader character-coding-error-external-format)))
