@@ -35,6 +35,7 @@
 
 (assert (zerop (deref (extern-alien "lowtag_for_widetag" (array char 64))
                       (ash sb-vm:character-widetag -2))))
+(gc :full t)
 
 ;;; Verify that all defstructs except for one were compiled in a null lexical
 ;;; environment. Compiling any call to a structure constructor would like to
@@ -69,7 +70,7 @@
           (destructuring-bind (fun args result) expr
             (let ((actual (apply fun (sb-int:ensure-list args))))
               (unless (eql actual result)
-                (#+sb-devel error
+                (#+sb-devel cerror #+sb-devel ""
                  #-sb-devel format #-sb-devel t
                  "FLOAT CACHE LINE ~S vs COMPUTED ~S~%"
                  expr actual)))))))))
