@@ -178,7 +178,7 @@
     (emit-label start-lab)
     ;; Skip space for the function header.
     (inst simple-fun-header-word)
-    (inst .skip (* (1- simple-fun-code-offset) n-word-bytes))
+    (inst .skip (* (1- simple-fun-insts-offset) n-word-bytes))
     ;; The start of the actual code.
     ;; Save the return-pc.
     (popw rbp-tn (frame-word-offset return-pc-save-offset))))
@@ -305,24 +305,30 @@
          (inst mov rsp-tn rbx-tn)))
       (t
        (collect ((defaults))
-         (let ((default-stack-slots (gen-label)))
+         (let ((default-stack-slots (gen-label))
+               (used-registers
+                 (loop for i from 1 below register-arg-count
+                       for tn = (tn-ref-tn (setf values (tn-ref-across values)))
+                       unless (eq (tn-kind tn) :unused)
+                       collect tn
+                       finally (setf values (tn-ref-across values))))
+               (used-stack-slots-p
+                 (loop for ref = values then (tn-ref-across ref)
+                       while ref
+                       thereis (neq (tn-kind (tn-ref-tn ref)) :unused))))
           (assemble ()
             (note-this-location vop :unknown-return)
-            ;; Branch off to the MV case.
-            (inst jmp :c regs-defaulted)
-            ;; Do the single value case.
-            ;; Default the register args
-            (do ((i 1 (1+ i))
-                 (val (tn-ref-across values) (tn-ref-across val)))
-                ((= i register-arg-count) (setf values val))
-              (unless (eq (tn-kind (tn-ref-tn val)) :unused)
-                (inst mov (tn-ref-tn val) nil-value)))
-            (move rbx-tn rsp-tn)
-            (if (loop for ref = values then (tn-ref-across ref)
-                      while ref
-                      thereis (neq (tn-kind (tn-ref-tn ref)) :unused))
-                (inst jmp default-stack-slots)
-                (inst jmp defaulting-done))
+            ;; If it returned exactly one value the registers and the
+            ;; stack slots need to be filled with NIL.
+            (cond (used-stack-slots-p
+                   (inst jmp :nc default-stack-slots))
+                  (t
+                   (inst jmp :c regs-defaulted)
+                   (loop for null = nil-value then (car used-registers)
+                         for reg in used-registers
+                         do (inst mov :dword reg null))
+                   (move rbx-tn rsp-tn)
+                   (inst jmp defaulting-done)))
             REGS-DEFAULTED
             (do ((i register-arg-count (1+ i))
                  (val values (tn-ref-across val)))
@@ -347,6 +353,10 @@
               (when defaults
                 (assemble (:elsewhere)
                   (emit-label default-stack-slots)
+                  (loop for null = nil-value then (car used-registers)
+                        for reg in used-registers
+                        do (inst mov :dword reg null))
+                  (move rbx-tn rsp-tn)
                   (dolist (default defaults)
                     (emit-label (car default))
                     (inst mov (cdr default) nil-value))
@@ -395,9 +405,9 @@
               register-arg-count)
       (inst cmp nargs (fixnumize register-arg-count))
       (inst jmp :g stack-values)
-      #!+#.(cl:if (cl:= sb-vm:word-shift sb-vm:n-fixnum-tag-bits) '(and) '(or))
+      #+#.(cl:if (cl:= sb-vm:word-shift sb-vm:n-fixnum-tag-bits) '(and) '(or))
       (inst sub rsp-tn nargs)
-      #!-#.(cl:if (cl:= sb-vm:word-shift sb-vm:n-fixnum-tag-bits) '(and) '(or))
+      #-#.(cl:if (cl:= sb-vm:word-shift sb-vm:n-fixnum-tag-bits) '(and) '(or))
       (progn
         ;; FIXME: This can't be efficient, but LEA (my first choice)
         ;; doesn't do subtraction.
@@ -604,7 +614,7 @@
 ;;; have been set up in the current frame.
 (macrolet ((define-full-call (vop-name named return variable)
             (aver (not (and variable (eq return :tail))))
-            #!+immobile-code (when named (setq named :direct))
+            #+immobile-code (when named (setq named :direct))
             `(define-vop (,vop-name ,@(when (eq return :unknown)
                                         '(unknown-values-receiver)))
                (:args
@@ -797,7 +807,7 @@
 
                (when (and step-instrumenting
                           ,@(and (eq named :direct)
-                                 `((not (and #!+immobile-code
+                                 `((not (and #+immobile-code
                                              ;; handle-single-step-around-trap can't handle it
                                              (static-fdefn-offset fun))))))
                  (emit-single-step-test)
@@ -806,7 +816,7 @@
                DONE
                (note-this-location vop :call-site)
                ,(cond ((eq named :direct)
-                       #!+immobile-code
+                       #+immobile-code
                        `(let* ((fixup (make-fixup
                                        fun
                                        (if (static-fdefn-offset fun)
@@ -826,10 +836,10 @@
                                        (inst mov rax-tn fixup)
                                        rax-tn))))
                           (inst ,(if (eq return :tail) 'jmp 'call) target))
-                       #!-immobile-code
+                       #-immobile-code
                        `(inst ,(if (eq return :tail) 'jmp 'call)
                               (ea (+ nil-value (static-fun-offset fun)))))
-                      #!-immobile-code
+                      #-immobile-code
                       (named
                        `(inst ,(if (eq return :tail) 'jmp 'call)
                               (ea (- (* fdefn-raw-addr-slot n-word-bytes)
@@ -849,15 +859,15 @@
 
   (define-full-call call nil :fixed nil)
   (define-full-call call-named t :fixed nil)
-  #!-immobile-code
+  #-immobile-code
   (define-full-call static-call-named :direct :fixed nil)
   (define-full-call multiple-call nil :unknown nil)
   (define-full-call multiple-call-named t :unknown nil)
-  #!-immobile-code
+  #-immobile-code
   (define-full-call static-multiple-call-named :direct :unknown nil)
   (define-full-call tail-call nil :tail nil)
   (define-full-call tail-call-named t :tail nil)
-  #!-immobile-code
+  #-immobile-code
   (define-full-call static-tail-call-named :direct :tail nil)
 
   (define-full-call call-variable nil :fixed t)
@@ -874,33 +884,33 @@
          (%lea-for-lowtag-test rbx-tn fun fun-pointer-lowtag)
          (inst test :byte rbx-tn lowtag-mask)
          (inst jmp :nz (if relative-call
-                           (make-fixup 'tail-call-symbol :assembly-routine)
+                           (make-fixup 'call-symbol :assembly-routine)
                            not-fun))
          (inst jmp fun-ea)
          not-fun
          (unless relative-call
-           (invoke-asm-routine 'jmp 'tail-call-symbol vop))))
+           (invoke-asm-routine 'jmp 'call-symbol vop))))
       (:symbol
-       (invoke-asm-routine 'jmp 'tail-call-symbol vop))
+       (invoke-asm-routine 'jmp 'call-symbol vop))
       (t
        (inst jmp fun-ea)))))
 
 (defun call-unnamed (fun type vop)
   (case type
     (:symbol
-     ;; CALL-SYMBOL returns past the CALL instruction that follows it,
-     ;; which is not needed here, hence TAIL-CALL-SYMBOL.
-     (invoke-asm-routine 'call 'tail-call-symbol vop))
+     (invoke-asm-routine 'call 'call-symbol vop))
     (t
      (assemble ()
        (when (eq type :designator)
          (%lea-for-lowtag-test rbx-tn fun fun-pointer-lowtag)
          (inst test :byte rbx-tn lowtag-mask)
          (inst jmp :z call)
-         (invoke-asm-routine 'call 'call-symbol vop))
+         (invoke-asm-routine 'call 'call-symbol vop)
+         (inst jmp ret))
        call
        (inst call (ea (- (* closure-fun-slot n-word-bytes) fun-pointer-lowtag)
-                      fun))))))
+                      fun))
+       ret))))
 
 ;;; This is defined separately, since it needs special code that BLT's
 ;;; the arguments down. All the real work is done in the assembly
@@ -1238,6 +1248,21 @@
     (inst mov value (ea object value
                         (ash 1 (- word-shift n-fixnum-tag-bits))))))
 
+(define-vop (more-arg-or-nil)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg) :to (:result 1))
+         (count :scs (any-reg) :to (:result 1)))
+  (:arg-types * tagged-num)
+  (:info index)
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:result-types *)
+  (:generator 3
+    (inst mov value nil-value)
+    (inst cmp count (fixnumize index))
+    (inst jmp :be done)
+    (inst mov value (ea (- (* index n-word-bytes)) object))
+    done))
+
 ;;; Turn more arg (context, count) into a list.
 (define-vop (listify-rest-args)
   (:translate %listify-rest-args)
@@ -1396,8 +1421,8 @@
   ;; and reading a slot from a thread structure would require an extra
   ;; register on -SB-THREAD. While this isn't critical for x86-64,
   ;; it's more serious for x86.
-  #!+sb-thread (inst cmp :byte (thread-slot-ea thread-stepping-slot) 0)
-  #!-sb-thread (inst cmp :byte (static-symbol-value-ea 'sb-impl::*stepping*) 0))
+  #+sb-thread (inst cmp :byte (thread-slot-ea thread-stepping-slot) 0)
+  #-sb-thread (inst cmp :byte (static-symbol-value-ea 'sb-impl::*stepping*) 0))
 
 (define-vop (step-instrument-before-vop)
   (:policy :fast-safe)
@@ -1407,4 +1432,4 @@
      (inst jmp :eq DONE)
      (inst break single-step-before-trap)
      DONE
-     (note-this-location vop :step-before-vop)))
+     (note-this-location vop :internal-error)))
