@@ -67,6 +67,7 @@
 #include "genesis/fdefn.h"
 #include "genesis/simple-fun.h"
 #include "genesis/cons.h"
+#include "genesis/vector.h"
 
 /*
  * This is a workaround for some slightly silly Linux/GNU Libc
@@ -76,12 +77,13 @@
  * it later into another ucontext: the ucontext is allocated on the
  * stack by the kernel, so copying a libc-sized sigset_t into it will
  * overflow and cause other data on the stack to be corrupted */
-/* FIXME: do not rely on NSIG being a multiple of 8 */
 /* See https://sourceware.org/bugzilla/show_bug.cgi?id=1780 */
 
 #ifdef LISP_FEATURE_WIN32
 # define REAL_SIGSET_SIZE_BYTES (4)
 #else
+/* FIXME: do not rely on NSIG being a multiple of 8.
+ * In fact it is *not* a multiple of 8 - it it 65 on x86-64-linux */
 # define REAL_SIGSET_SIZE_BYTES ((NSIG/8))
 #endif
 
@@ -98,6 +100,8 @@ sigcopyset(sigset_t *new, sigset_t *old)
  * becomes 'yes'.) */
 boolean internal_errors_enabled = 0;
 
+// SIGRTMAX is not usable in an array size declaration because it might be
+// a variable expression, so use NSIG which is at least as large as SIGRTMAX.
 #ifndef LISP_FEATURE_WIN32
 static
 void (*interrupt_low_level_handlers[NSIG]) (int, siginfo_t*, os_context_t*);
@@ -116,6 +120,49 @@ union interrupt_handler interrupt_handlers[NSIG];
     os_context_t *context = arch_os_get_context(&void_context);
 #endif
 
+/* The code below wants to iterate from 1 through SIGRTMAX inclusive, not NSIG,
+ * because NSIG is the maximum capacity of a sigset, whereas SIGRTMAX is a
+ * possibly smaller value indicating the number of signal numbers that could
+ * really be in use. For systems which have distinctly different values
+ * (discounting the deliberate off-by-one nature), sigismember() may return -1
+ * for signal numbers in excess of SIGRTMAX, and we would generally take
+ * -1 to mean boolean 'true' unless it is carefully checked for.
+ * The following program illustrates the issue:
+ *    #include <signal.h>
+ *    #include <stdio.h>
+ *    void main() {
+ *      sigset_t ss;
+ *      sigemptyset(&ss);
+ *      printf("%d %d %d\n", SIGRTMAX, NSIG, sigismember(&ss, 41));
+ *    }
+ *  ./sigismembertest => 40 65 -1
+ */
+#ifdef SIGRTMAX
+#  define MAX_SIGNUM SIGRTMAX
+#else
+#  define MAX_SIGNUM (NSIG-1)
+#endif
+
+// For each bit present in 'source', add it to 'dest'.
+// This is the same as "sigorset(dest, dest, source)" if _GNU_SOURCE is defined.
+static void sigmask_logior(sigset_t *dest, const sigset_t *source)
+{
+    int i;
+    for(i = 1; i <= MAX_SIGNUM; i++) {
+        if (sigismember(source, i)) sigaddset(dest, i);
+    }
+}
+
+// For each bit present in 'source', remove it from 'dest'
+// (logically AND with logical complement).
+static void sigmask_logandc(sigset_t *dest, const sigset_t *source)
+{
+    int i;
+    for(i = 1; i <= MAX_SIGNUM; i++) {
+        if (sigismember(source, i)) sigdelset(dest, i);
+    }
+}
+
 /* Foreign code may want to start some threads on its own.
  * Non-targetted, truly asynchronous signals can be delivered to
  * basically any thread, but invoking Lisp handlers in such foregign
@@ -137,8 +184,6 @@ add_handled_signals(sigset_t *sigset)
         }
     }
 }
-
-void block_signals(sigset_t *what, sigset_t *where, sigset_t *old);
 #endif
 
 static boolean
@@ -156,14 +201,15 @@ maybe_resignal_to_lisp_thread(int signal, os_context_t *context)
              signal);
 #endif
         }
-        {
-            sigset_t sigset;
-            sigemptyset(&sigset);
-            add_handled_signals(&sigset);
-            block_signals(&sigset, 0, 0);
-            block_signals(&sigset, os_context_sigmask_addr(context), 0);
-            kill(getpid(), signal);
-        }
+        sigset_t sigset;
+        sigemptyset(&sigset);
+        add_handled_signals(&sigset);
+        thread_sigmask(SIG_BLOCK, &sigset, 0);
+        // This arranges for every handled signal to be blocked on return from this
+        // handler invocation, presumably because that avoids further detours through
+        // this thread's handler for signals that we don't want it to handle.
+        sigmask_logior(os_context_sigmask_addr(context), &sigset);
+        kill(getpid(), signal);
         return 1;
     } else
 #endif
@@ -247,45 +293,13 @@ get_current_sigmask(sigset_t *sigset)
     thread_sigmask(SIG_BLOCK, 0, sigset);
 }
 
-void
-block_signals(sigset_t *what, sigset_t *where, sigset_t *old)
-{
-    if (where) {
-        int i;
-        if (old)
-            sigcopyset(old, where);
-        for(i = 1; i < NSIG; i++) {
-            if (sigismember(what, i))
-                sigaddset(where, i);
-        }
-    } else {
-        thread_sigmask(SIG_BLOCK, what, old);
-    }
-}
-
-void
-unblock_signals(sigset_t *what, sigset_t *where, sigset_t *old)
-{
-    if (where) {
-        int i;
-        if (old)
-            sigcopyset(old, where);
-        for(i = 1; i < NSIG; i++) {
-            if (sigismember(what, i))
-                sigdelset(where, i);
-        }
-    } else {
-        thread_sigmask(SIG_UNBLOCK, what, old);
-    }
-}
-
 // Stringify sigset into the supplied result buffer.
 static void
-sigset_tostring(sigset_t *sigset, char* result, int result_length)
+sigset_tostring(const sigset_t *sigset, char* result, int result_length)
 {
     int i;
     int len = 0;
-    for(i = 1; i < NSIG; i++)
+    for(i = 1; i <= MAX_SIGNUM; i++)
         if (sigismember(sigset, i)) {
             // ensure room for (generously) 3 digits + comma + null, or give up
             if (len > result_length - 5) {
@@ -297,12 +311,12 @@ sigset_tostring(sigset_t *sigset, char* result, int result_length)
     result[len] = 0;
 }
 
-/* Return 1 is all signals is sigset2 are masked in sigset, return 0
- * if all re unmasked else die. Passing NULL for sigset is a shorthand
+/* Return 1 if all signals in sigset2 are masked in sigset, return 0
+ * if all are unmasked, else die. Passing NULL for sigset is a shorthand
  * for the current sigmask. */
 boolean
-all_signals_blocked_p(sigset_t *sigset, sigset_t *sigset2,
-                                const char *name)
+all_signals_blocked_p(const sigset_t *sigset, sigset_t *sigset2,
+                      const char *name)
 {
     int i;
     boolean has_blocked = 0, has_unblocked = 0;
@@ -311,7 +325,7 @@ all_signals_blocked_p(sigset_t *sigset, sigset_t *sigset2,
         get_current_sigmask(&current);
         sigset = &current;
     }
-    for(i = 1; i < NSIG; i++) {
+    for(i = 1; i <= MAX_SIGNUM; i++) {
         if (sigismember(sigset2, i)) {
             if (sigismember(sigset, i))
                 has_blocked = 1;
@@ -322,7 +336,7 @@ all_signals_blocked_p(sigset_t *sigset, sigset_t *sigset2,
     if (has_blocked && has_unblocked) {
         char buf[3*64]; // assuming worst case 64 signals present in sigset
         sigset_tostring(sigset, buf, sizeof buf);
-        lose("%s signals partially blocked: {%s}\n", name);
+        lose("%s signals partially blocked: {%s}\n", name, buf);
     }
     if (has_blocked)
         return 1;
@@ -345,7 +359,11 @@ sigaddset_deferrable(sigset_t *s)
     sigaddset(s, SIGURG);
     sigaddset(s, SIGTSTP);
     sigaddset(s, SIGCHLD);
+#ifdef SIGIO
     sigaddset(s, SIGIO);
+#else
+    sigaddset(s, SIGPOLL);
+#endif
 #ifndef LISP_FEATURE_HPUX
     sigaddset(s, SIGXCPU);
     sigaddset(s, SIGXFSZ);
@@ -363,7 +381,7 @@ sigaddset_blockable(sigset_t *sigset)
 }
 
 void
-sigaddset_gc(sigset_t *sigset)
+sigaddset_gc(sigset_t __attribute__((unused)) *sigset)
 {
 #ifdef THREADS_USING_GCSIGNAL
     sigaddset(sigset,SIG_STOP_FOR_GC);
@@ -476,7 +494,7 @@ void
 block_deferrable_signals(sigset_t *old)
 {
 #if !defined(LISP_FEATURE_WIN32) || defined(LISP_FEATURE_SB_THREAD)
-    block_signals(&deferrable_sigset, 0, old);
+    thread_sigmask(SIG_BLOCK, &deferrable_sigset, old);
 #endif
 }
 
@@ -484,30 +502,38 @@ void
 block_blockable_signals(sigset_t *old)
 {
 #if !defined(LISP_FEATURE_WIN32) || defined(LISP_FEATURE_SB_THREAD)
-    block_signals(&blockable_sigset, 0, old);
+    thread_sigmask(SIG_BLOCK, &blockable_sigset, old);
 #endif
 }
 
+// Do one of two things depending on whether the specified 'where'
+// is non-null or null.
+// 1. If non-null, then alter the mask in *where, removing deferrable signals
+//    from it without affecting the thread's mask from perspective of the OS.
+// 2. If null, the actually change the current thread's signal mask.
 void
-unblock_deferrable_signals(sigset_t *where, sigset_t *old)
+unblock_deferrable_signals(sigset_t *where)
 {
 #if !defined(LISP_FEATURE_WIN32) || defined(LISP_FEATURE_SB_THREAD)
     if (interrupt_handler_pending_p())
         lose("unblock_deferrable_signals: losing proposition\n");
 #ifndef LISP_FEATURE_SB_SAFEPOINT
+    // If 'where' is null, check_gc_signals_unblocked_or_lose() will
+    // fetch the current signal mask (from the OS) and check that.
     check_gc_signals_unblocked_or_lose(where);
 #endif
-    unblock_signals(&deferrable_sigset, where, old);
+    if (where)
+        sigmask_logandc(where, &deferrable_sigset);
+    else
+        thread_sigmask(SIG_UNBLOCK, &deferrable_sigset, 0);
 #endif
 }
 
 #ifndef LISP_FEATURE_SB_SAFEPOINT
-void
-unblock_gc_signals(sigset_t *where, sigset_t *old)
-{
-#ifndef LISP_FEATURE_WIN32
-    unblock_signals(&gc_sigset, where, old);
-#endif
+// This function previously had an #ifdef guard precluding doing anything for
+// win32, which was redundant because SB_SAFEPOINT is always defined for win32.
+void unblock_gc_signals(void) {
+    thread_sigmask(SIG_UNBLOCK, &gc_sigset, 0);
 }
 #endif
 
@@ -522,11 +548,11 @@ unblock_signals_in_context_and_maybe_warn(os_context_t *context)
 "Enabling blocked gc signals to allow returning to Lisp without risking\n\
 gc deadlocks. Since GC signals are only blocked in signal handlers when \n\
 they are not safe to interrupt at all, this is a pretty severe occurrence.\n");
-        unblock_gc_signals(sigset, 0);
+        sigmask_logandc(sigset, &gc_sigset);
     }
 #endif
     if (!interrupt_handler_pending_p()) {
-        unblock_deferrable_signals(sigset, 0);
+        unblock_deferrable_signals(sigset);
     }
 #endif
 }
@@ -592,7 +618,7 @@ maybe_save_gc_mask_and_block_deferrables(sigset_t *sigset)
  * enabled, without the protection of *GC-INHIBIT* T and there is gc
  * (or stop for gc) pending, but we haven't trapped yet? */
 int
-in_leaving_without_gcing_race_p(struct thread *thread)
+in_leaving_without_gcing_race_p(struct thread __attribute__((unused)) *thread)
 {
     return ((read_TLS(IN_WITHOUT_GCING,thread) != NIL) &&
             (read_TLS(INTERRUPTS_ENABLED,thread) != NIL) &&
@@ -688,7 +714,8 @@ check_interrupt_context_or_lose(os_context_t *context)
  */
 
 static void
-build_fake_control_stack_frames(struct thread *th,os_context_t *context)
+build_fake_control_stack_frames(struct thread __attribute__((unused)) *th,
+                                os_context_t __attribute__((unused)) *context)
 {
 #ifndef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
 
@@ -759,7 +786,7 @@ fake_foreign_function_call(os_context_t *context)
     check_blockables_blocked_or_lose(0);
 
     /* Get current Lisp state from context. */
-#if defined(LISP_FEATURE_ARM) && !defined(LISP_FEATURE_GENCGC)
+#if (defined(LISP_FEATURE_ARM) || defined(LISP_FEATURE_RISCV)) && !defined(LISP_FEATURE_GENCGC)
     dynamic_space_free_pointer = SymbolValue(ALLOCATION_POINTER, thread);
 #endif
 #ifdef reg_ALLOC
@@ -803,8 +830,9 @@ fake_foreign_function_call(os_context_t *context)
     context_index =
         fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX,thread));
 
-    if (context_index >= MAX_INTERRUPTS) {
-        lose("maximum interrupt nesting depth (%d) exceeded\n", MAX_INTERRUPTS);
+    if (context_index >= (MAX_INTERRUPTS-THREAD_HEADER_SLOTS)) {
+        lose("maximum interrupt nesting depth (%d) exceeded\n",
+             MAX_INTERRUPTS-THREAD_HEADER_SLOTS);
     }
 
     bind_variable(FREE_INTERRUPT_CONTEXT_INDEX,
@@ -824,7 +852,7 @@ fake_foreign_function_call(os_context_t *context)
  * the usual signal mask will be restored from the context when the handler
  * finishes.  Otherwise, be careful */
 void
-undo_fake_foreign_function_call(os_context_t *context)
+undo_fake_foreign_function_call(os_context_t __attribute__((unused)) *context)
 {
     struct thread *thread=arch_os_get_current_thread();
     /* Block all blockable signals. */
@@ -873,7 +901,7 @@ undo_fake_foreign_function_call(os_context_t *context)
      * not updating them. */
     thread->pseudo_atomic_bits = 0;
 #endif
-#if defined(LISP_FEATURE_ARM) && !defined(LISP_FEATURE_GENCGC)
+#if (defined(LISP_FEATURE_ARM) || defined (LISP_FEATURE_RISCV)) && !defined(LISP_FEATURE_GENCGC)
     SetSymbolValue(ALLOCATION_POINTER, dynamic_space_free_pointer, thread);
 #endif
 }
@@ -894,13 +922,7 @@ interrupt_internal_error(os_context_t *context, boolean continuable)
         lose("internal error too early in init, can't recover\n");
     }
 
-#ifndef LISP_FEATURE_SB_SAFEPOINT
-    unblock_gc_signals(0, 0);
-#endif
-
-#if !defined(LISP_FEATURE_WIN32) || defined(LISP_FEATURE_SB_THREAD)
     thread_sigmask(SIG_SETMASK, os_context_sigmask_addr(context), 0);
-#endif
 
 #if defined(LISP_FEATURE_LINUX) && defined(LISP_FEATURE_MIPS)
     /* Workaround for blocked SIGTRAP. */
@@ -1079,7 +1101,7 @@ interrupt_handle_pending(os_context_t *context)
          * an interrupt arrived during GC (POST-GC, really) it was
          * handled. */
         if (original_pending_handler != data->pending_handler)
-            lose("pending handler changed in gc: %x -> %x.",
+            lose("pending handler changed in gc: %p -> %p.",
                  original_pending_handler, data->pending_handler);
     }
 
@@ -1157,27 +1179,22 @@ interrupt_handle_now(int signal, siginfo_t *info, os_context_t *context)
          * support decides to pass on it. */
         lose("no handler for signal %d in interrupt_handle_now(..)\n", signal);
 
+        // BUG: if a C function pointer can be misaligned such that it
+        // looks to satisfy functionp() then we do the wrong thing.
     } else if (functionp(handler.lisp)) {
         /* Once we've decided what to do about contexts in a
          * return-elsewhere world (the original context will no longer
          * be available; should we copy it or was nobody using it anyway?)
          * then we should convert this to return-elsewhere */
 
-        /* CMUCL comment said "Allocate the SAPs while the interrupts
-         * are still disabled.".  I (dan, 2003.08.21) assume this is
-         * because we're not in pseudoatomic and allocation shouldn't
-         * be interrupted.  In which case it's no longer an issue as
-         * all our allocation from C now goes through a PA wrapper,
-         * but still, doesn't hurt.
-         *
-         * Yeah, but non-gencgc platforms don't really wrap allocation
-         * in PA. MG - 2005-08-29  */
-
-
 #ifndef LISP_FEATURE_SB_SAFEPOINT
         /* Leave deferrable signals blocked, the handler itself will
          * allow signals again when it sees fit. */
-        unblock_gc_signals(0, 0);
+#ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
+        /* handler.lisp will hide from the GC, will be enabled in the handler itself.
+         * Not a problem for the conservative GC. */
+        unblock_gc_signals();
+#endif
 #else
         WITH_GC_AT_SAFEPOINTS_ONLY()
 #endif
@@ -1261,8 +1278,8 @@ maybe_defer_handler(void *handler, struct interrupt_data *data,
     if ((read_TLS(INTERRUPTS_ENABLED,thread) == NIL) ||
         in_leaving_without_gcing_race_p(thread)) {
         FSHOW_SIGNAL((stderr,
-                      "/maybe_defer_handler(%x,%d): deferred (RACE=%d)\n",
-                      (unsigned int)handler,signal,
+                      "/maybe_defer_handler(%p,%d): deferred (RACE=%d)\n",
+                      handler,signal,
                       in_leaving_without_gcing_race_p(thread)));
         store_signal_data_for_later(data,handler,signal,info,context);
         write_TLS(INTERRUPT_PENDING, T,thread);
@@ -1274,8 +1291,8 @@ maybe_defer_handler(void *handler, struct interrupt_data *data,
      * may succeed even when context is null (gencgc alloc()) */
     if (arch_pseudo_atomic_atomic(context)) {
         FSHOW_SIGNAL((stderr,
-                      "/maybe_defer_handler(%x,%d): deferred(PA)\n",
-                      (unsigned int)handler,signal));
+                      "/maybe_defer_handler(%p,%d): deferred(PA)\n",
+                      handler,signal));
         store_signal_data_for_later(data,handler,signal,info,context);
         arch_set_pseudo_atomic_interrupted(context);
         check_interrupt_context_or_lose(context);
@@ -1285,8 +1302,8 @@ maybe_defer_handler(void *handler, struct interrupt_data *data,
     check_interrupt_context_or_lose(context);
 
     FSHOW_SIGNAL((stderr,
-                  "/maybe_defer_handler(%x,%d): not deferred\n",
-                  (unsigned int)handler,signal));
+                  "/maybe_defer_handler(%p,%d): not deferred\n",
+                  handler,signal));
     return 0;
 }
 
@@ -1296,7 +1313,7 @@ store_signal_data_for_later (struct interrupt_data *data, void *handler,
                              siginfo_t *info, os_context_t *context)
 {
     if (data->pending_handler)
-        lose("tried to overwrite pending interrupt handler %x with %x\n",
+        lose("tried to overwrite pending interrupt handler %p with %p",
              data->pending_handler, handler);
     if (!handler)
         lose("tried to defer null interrupt handler\n");
@@ -1359,7 +1376,9 @@ low_level_maybe_now_maybe_later(int signal, siginfo_t *info, void *void_context)
 
 /* This function must not cons, because that may trigger a GC. */
 void
-sig_stop_for_gc_handler(int signal, siginfo_t *info, os_context_t *context)
+sig_stop_for_gc_handler(int __attribute__((unused)) signal,
+                        siginfo_t __attribute__((unused)) *info,
+                        os_context_t *context)
 {
     struct thread *thread=arch_os_get_current_thread();
     boolean was_in_lisp;
@@ -1412,8 +1431,10 @@ sig_stop_for_gc_handler(int signal, siginfo_t *info, os_context_t *context)
     }
 
     if(thread_state(thread)!=STATE_RUNNING) {
-        lose("sig_stop_for_gc_handler: wrong thread state: %ld\n",
-             fixnum_value(thread->state));
+        // macOS warns if OBJ_FMTX is used to format a 'sword_t'
+        // which fixnum_value() returns.
+        lose("sig_stop_for_gc_handler: wrong thread state: %"OBJ_FMTX,
+             (lispobj)fixnum_value(thread->state));
     }
 
     set_thread_state(thread,STATE_STOPPED);
@@ -1429,8 +1450,8 @@ sig_stop_for_gc_handler(int signal, siginfo_t *info, os_context_t *context)
     FSHOW_SIGNAL((stderr,"resumed\n"));
 
     if(thread_state(thread)!=STATE_RUNNING) {
-        lose("sig_stop_for_gc_handler: wrong thread state on wakeup: %ld\n",
-             fixnum_value(thread_state(thread)));
+        lose("sig_stop_for_gc_handler: wrong thread state on wakeup: %"OBJ_FMTX,
+             (lispobj)fixnum_value(thread_state(thread)));
     }
 
     if (was_in_lisp) {
@@ -1446,7 +1467,7 @@ interrupt_handle_now_handler(int signal, siginfo_t *info, void *void_context)
     SAVE_ERRNO(signal,context,void_context);
 #ifndef LISP_FEATURE_WIN32
     if ((signal == SIGILL) || (signal == SIGBUS)
-#if !(defined(LISP_FEATURE_LINUX) || defined(LISP_FEATURE_ANDROID))
+#if !(defined LISP_FEATURE_LINUX || defined LISP_FEATURE_ANDROID || defined LISP_FEATURE_HAIKU)
         || (signal == SIGEMT)
 #endif
         )
@@ -1466,7 +1487,11 @@ interrupt_handle_now_handler(int signal, siginfo_t *info, void *void_context)
 extern int *context_eflags_addr(os_context_t *context);
 #endif
 
+#ifdef CALL_INTO_LISP
+#define call_into_lisp ((lispobj(*)(lispobj fun, lispobj *args, int nargs))SYMBOL(CALL_INTO_LISP)->value)
+#else
 extern lispobj call_into_lisp(lispobj fun, lispobj *args, int nargs);
+#endif
 extern void post_signal_tramp(void);
 extern void call_into_lisp_tramp(void);
 
@@ -1481,7 +1506,7 @@ arrange_return_to_c_function(os_context_t *context,
 #endif
 #if !(defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64))
     void * fun=native_pointer(function);
-    void *code = &(((struct simple_fun *) fun)->code);
+    void *code = &(((struct simple_fun *) fun)->insts);
 #endif
 
     /* Build a stack frame showing `interrupted' so that the
@@ -1653,7 +1678,7 @@ arrange_return_to_c_function(os_context_t *context,
     *os_context_npc_addr(context) =
         4 + *os_context_pc_addr(context);
 #endif
-#if defined(LISP_FEATURE_SPARC) || defined(LISP_FEATURE_ARM) || defined(LISP_FEATURE_ARM64)
+#if defined(LISP_FEATURE_SPARC) || defined(LISP_FEATURE_ARM) || defined(LISP_FEATURE_ARM64) || defined(LISP_FEATURE_RISCV)
     *os_context_register_addr(context,reg_CODE) =
         (os_context_register_t)((char*)fun + FUN_POINTER_LOWTAG);
 #endif
@@ -1723,7 +1748,7 @@ handle_guard_page_triggered(os_context_t *context,os_vm_address_t addr)
        addr < CONTROL_STACK_HARD_GUARD_PAGE(th) + os_vm_page_size) {
         fake_foreign_function_call(context);
         lose("Control stack exhausted, fault: %p, PC: %p",
-             addr, *os_context_pc_addr(context));
+             addr, (void*)*os_context_pc_addr(context));
     }
     else if(addr >= CONTROL_STACK_GUARD_PAGE(th) &&
             addr < CONTROL_STACK_GUARD_PAGE(th) + os_vm_page_size) {
@@ -1731,10 +1756,20 @@ handle_guard_page_triggered(os_context_t *context,os_vm_address_t addr)
          * protection so the error handler has some headroom, protect the
          * previous page so that we can catch returns from the guard page
          * and restore it. */
-        if (lose_on_corruption_p || gc_active_p) {
+        if (gc_active_p) {
+            fake_foreign_function_call(context);
+            lose("Control stack exhausted with gc_active_p, fault: %p, PC: %p",
+                 addr, (void*)*os_context_pc_addr(context));
+        }
+        if (arch_pseudo_atomic_atomic(context)) {
+            fake_foreign_function_call(context);
+            lose("Control stack exhausted while pseudo-atomic, fault: %p, PC: %p",
+                 addr, (void*)*os_context_pc_addr(context));
+        }
+        if (lose_on_corruption_p) {
             fake_foreign_function_call(context);
             lose("Control stack exhausted, fault: %p, PC: %p",
-                 addr, *os_context_pc_addr(context));
+                 addr, (void*)*os_context_pc_addr(context));
         }
         if (th->control_stack_guard_page_protected == NIL)
             lose("control_stack_guard_page_protected NIL");
@@ -1846,7 +1881,9 @@ static volatile int sigaction_nodefer_works = -1;
 #define SA_NODEFER_TEST_KILL_SIGNAL SIGUSR1
 
 static void
-sigaction_nodefer_test_handler(int signal, siginfo_t *info, void *void_context)
+sigaction_nodefer_test_handler(int signal,
+                               siginfo_t __attribute__((unused)) *info,
+                               void __attribute__((unused)) *void_context)
 {
     sigset_t current;
     int i;
@@ -1856,7 +1893,7 @@ sigaction_nodefer_test_handler(int signal, siginfo_t *info, void *void_context)
      * the signal we're handling when SA_NODEFER is set; Linux before
      * 2.6.13 or so also doesn't block the other signal when
      * SA_NODEFER is set. */
-    for(i = 1; i < NSIG; i++)
+    for(i = 1; i <= MAX_SIGNUM; i++)
         if (sigismember(&current, i) !=
             (((i == SA_NODEFER_TEST_BLOCK_SIGNAL) || (i == signal)) ? 1 : 0)) {
             FSHOW_SIGNAL((stderr, "SA_NODEFER doesn't work, signal %d\n", i));
@@ -1891,6 +1928,16 @@ see_if_sigaction_nodefer_works(void)
 #undef SA_NODEFER_TEST_BLOCK_SIGNAL
 #undef SA_NODEFER_TEST_KILL_SIGNAL
 
+extern void restore_sbcl_signals () {
+    int signal;
+    for (signal = 0; signal < NSIG; signal++) {
+        interrupt_handler_t handler = interrupt_low_level_handlers[signal];
+        if (handler) {
+            undoably_install_low_level_interrupt_handler(signal, handler);
+        }
+    }
+}
+
 #if defined(LISP_FEATURE_SB_SAFEPOINT_STRICTLY) && !defined(LISP_FEATURE_WIN32)
 
 static void *
@@ -1899,13 +1946,13 @@ signal_thread_trampoline(void *pthread_arg)
     intptr_t signo = (intptr_t) pthread_arg;
     os_context_t fake_context;
     siginfo_t fake_info;
-#ifdef LISP_FEATURE_PPC
+#if defined LISP_FEATURE_PPC || defined LISP_FEATURE_PPC64
     mcontext_t uc_regs;
 #endif
 
     memset(&fake_info, 0, sizeof(fake_info));
     memset(&fake_context, 0, sizeof(fake_context));
-#ifdef LISP_FEATURE_PPC
+#if defined LISP_FEATURE_PPC || defined LISP_FEATURE_PPC64
     memset(&uc_regs, 0, sizeof(uc_regs));
     fake_context.uc_mcontext.uc_regs = &uc_regs;
 #endif
@@ -2059,12 +2106,12 @@ undoably_install_low_level_interrupt_handler (int signal,
 /* This is called from Lisp. */
 uword_t
 install_handler(int signal, void handler(int, siginfo_t*, os_context_t*),
-                int synchronous)
+                lispobj ohandler, // (SIMPLE-VECTOR 1) as a tagged pointer
+                int __attribute__((unused)) synchronous)
 {
 #ifndef LISP_FEATURE_WIN32
     struct sigaction sa;
     sigset_t old;
-    union interrupt_handler oldhandler;
 
     FSHOW((stderr, "/entering POSIX install_handler(%d, ..)\n", signal));
 
@@ -2073,9 +2120,10 @@ install_handler(int signal, void handler(int, siginfo_t*, os_context_t*),
     FSHOW((stderr, "/interrupt_low_level_handlers[signal]=%p\n",
            interrupt_low_level_handlers[signal]));
     if (interrupt_low_level_handlers[signal]==0) {
-        if (ARE_SAME_HANDLER(handler, SIG_DFL) ||
-            ARE_SAME_HANDLER(handler, SIG_IGN))
-            sa.sa_sigaction = (void (*)(int, siginfo_t*, void*))handler;
+        if (handler == 0)
+            sa.sa_sigaction = (void (*)(int, siginfo_t*, void*))SIG_DFL;
+        else if ((lispobj)handler == 1)
+            sa.sa_sigaction = (void (*)(int, siginfo_t*, void*))SIG_IGN;
 #ifdef LISP_FEATURE_SB_SAFEPOINT_STRICTLY
         else if (signal == SIGPROF)
             sa.sa_sigaction = sigprof_handler_trampoline;
@@ -2096,14 +2144,37 @@ install_handler(int signal, void handler(int, siginfo_t*, os_context_t*),
         sigaction(signal, &sa, NULL);
     }
 
-    oldhandler = interrupt_handlers[signal];
+    // We can GC-safely read interrupt_low_level_handlers[signal] and assign
+    // into ohandler despite that the C stack and registers are not roots on
+    // precise GC. That is, if GC were to occur as depicted:
+    //   obj = old_handler
+    //   -- GC signal occurs here --
+    //   result = obj
+    // then the register or stack word holding 'obj' could be obsolete.
+    // But that won't happen- the STOP_FOR_GC signal is currently blocked
+    // due to the block_blockable_signals() at the top of this function.
+    // However, there is an unlikely other reason for a wrong result -
+    // if the C function address does not have at least N_FIXNUM_TAG_BITS
+    // of alignment, then we say that it was NIL instead of a fixnum.
+    lispobj obj = interrupt_handlers[signal].lisp;
+    uword_t result = -1;
+    if ((void*)obj == (void*)SIG_DFL)
+        result = 0;
+    else if ((void*)obj == (void*)SIG_IGN)
+        result = 1;
+    else if (ohandler) // only store into 'ohandler' when it was supplied
+        VECTOR(ohandler)->data[0] = (functionp(obj)||fixnump(obj)) ? obj : NIL;
+
     interrupt_handlers[signal].c = handler;
 
     thread_sigmask(SIG_SETMASK, &old, 0);
 
     FSHOW((stderr, "/leaving POSIX install_handler(%d, ..)\n", signal));
 
-    return (uword_t)oldhandler.lisp;
+    // After the thread's signal mask is restored to the mask in 'old',
+    // it is not possible to GC-safely refer to a Lisp function from C code,
+    // which is to say, naively utilizing "return obj;" would be wrong.
+    return result;
 #else
     /* Probably-wrong Win32 hack */
     return 0;
@@ -2113,7 +2184,9 @@ install_handler(int signal, void handler(int, siginfo_t*, os_context_t*),
 /* This must not go through lisp as it's allowed anytime, even when on
  * the altstack. */
 void
-sigabrt_handler(int signal, siginfo_t *info, os_context_t *context)
+sigabrt_handler(int __attribute__((unused)) signal,
+                siginfo_t __attribute__((unused)) *info,
+                os_context_t *context)
 {
     /* Save the interrupt context. No need to undo it, since lose()
      * shouldn't return. */
@@ -2168,16 +2241,27 @@ lisp_memory_fault_error(os_context_t *context, os_vm_address_t addr)
     fake_foreign_function_call(context);
 
     /* To allow debugging memory faults in signal handlers and such. */
-    corruption_warning_and_maybe_lose("Memory fault at %p (pc=%p, sp=%p)",
+#ifdef ARCH_HAS_STACK_POINTER
+    corruption_warning_and_maybe_lose("Memory fault at %p (pc=%p, fp=%p, sp=%p) tid %#lx",
                                       addr,
                                       *os_context_pc_addr(context),
-#ifdef ARCH_HAS_STACK_POINTER
-                                      *os_context_sp_addr(context)
+                                      os_context_frame_pointer(context),
+                                      *os_context_sp_addr(context),
+                                      thread_self()); // = 0 if -sb-thread
 #else
-                                      0
+    corruption_warning_and_maybe_lose("Memory fault at %p (pc=%p)",
+                                      addr, *os_context_pc_addr(context));
 #endif
-                                      );
+
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
+    /* Holy hell is this more obfuscated than necessary when using
+     * signal emulation on macOS. It almost makes one want to cry.
+     * We're not actually on an alternate stack at this point.
+     * Instead of telling the emulated sigsegv (which needn't have been
+     * emulated at all) to return to an intruction which executes a
+     * sigtrap (also emulated), we should just go straight where
+     * we need to go and hand it the original context rather than
+     * having to track the context through two bogus signals */
 #  if !(defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64))
 #    error memory fault emulation needs validating for this architecture
 #  endif
@@ -2229,9 +2313,9 @@ handle_memory_fault_emulation_trap(os_context_t *context)
 #endif /* C_STACK_IS_CONTROL_STACK */
     /* On x86oids, we're in handle_memory_fault_emulation_trap().
      * On real computers, we're still in lisp_memory_fault_error(). */
-#ifndef LISP_FEATURE_SB_SAFEPOINT
-    unblock_gc_signals(0, 0);
-#endif
+
+    thread_sigmask(SIG_SETMASK, os_context_sigmask_addr(context), 0);
+
     DX_ALLOC_SAP(context_sap, context);
     DX_ALLOC_SAP(fault_address_sap, addr);
     funcall2(StaticSymbolFunction(MEMORY_FAULT_ERROR),
@@ -2245,13 +2329,9 @@ unhandled_trap_error(os_context_t *context)
 {
     DX_ALLOC_SAP(context_sap, context);
     fake_foreign_function_call(context);
-#ifndef LISP_FEATURE_SB_SAFEPOINT
-    unblock_gc_signals(0, 0);
-#endif
 
-#if !defined(LISP_FEATURE_WIN32) || defined(LISP_FEATURE_SB_THREAD)
     thread_sigmask(SIG_SETMASK, os_context_sigmask_addr(context), 0);
-#endif
+
     funcall1(StaticSymbolFunction(UNHANDLED_TRAP_ERROR), context_sap);
     lose("UNHANDLED-TRAP-ERROR fell through");
 }
