@@ -2382,7 +2382,7 @@
 
 (with-test (:name (:mv-call :more-arg-unused)
             ;; needs SB-VM::MORE-ARG-OR-NIL VOP
-            :skipped-on (not (or :x86-64 :x86 :ppc :arm :arm64 :riscv)))
+            :broken-on (not (or :x86-64 :x86 :ppc :arm :arm64 :riscv)))
   (checked-compile-and-assert
    ()
    '(lambda (&rest rest)
@@ -2391,3 +2391,61 @@
    (() '(nil nil) :test #'equal)
    ((1) '(1 nil) :test #'equal)
    ((1 3) '(1 3) :test #'equal)))
+
+(with-test (:name :truncate-deriver-on-number-type)
+  (checked-compile-and-assert
+   ()
+   '(lambda (i)
+     (truncate
+      (labels ((f (&optional (o i))
+                 (declare (ignore o))
+                 (complex 0 0)))
+        (declare (dynamic-extent (function f)))
+        (the integer
+             (multiple-value-call #'f (values))))
+      3))
+   ((0) (values 0 0))))
+
+(with-test (:name :signum-type-deriver)
+  (checked-compile-and-assert
+   ()
+   '(lambda (n)
+     (typep (signum n) 'complex))
+   ((#C(1 2)) t)
+   ((1d0) nil)
+   ((10) nil)))
+
+(with-test (:name :array-header-p-derivation)
+  (checked-compile-and-assert
+   ()
+   '(lambda (q)
+     (and (typep q '(not simple-array))
+      (sb-kernel:array-header-p q)))
+   ((10) nil)
+   (((make-array 10 :adjustable t)) t)))
+
+(with-test (:name :phase-type-derivation)
+  (checked-compile-and-assert
+   ()
+   '(lambda (x)
+     (= (phase (the (integer -1 0) x))
+      (coerce pi 'single-float)))
+   ((-1) t)
+   ((0) nil)))
+
+(with-test (:name :maybe-negate-check-fun-type)
+  (checked-compile-and-assert
+   ()
+   '(lambda (m)
+     (declare ((or (function (number)) (eql #.#'symbolp)) m))
+     (the (member 3/4 4/5 1/2 #.#'symbolp) m))
+   ((#'symbolp) #'symbolp)))
+
+(with-test (:name :lvar-fun-type-on-literal-funs)
+  (checked-compile-and-assert
+   ()
+   `(lambda (p)
+      (declare (type (or null string) p))
+      (locally (declare (optimize (space 0)))
+        (stable-sort p ,#'string<)))
+   (((copy-seq "acb")) "abc" :test #'equal)))
